package eu.ztsh.garmin

import android.annotation.SuppressLint
import android.bluetooth.BluetoothAdapter
import android.bluetooth.BluetoothDevice
import android.bluetooth.BluetoothSocket
import android.util.Log
import com.mapbox.navigation.tripdata.maneuver.model.Maneuver
import com.mapbox.navigation.core.trip.session.LocationMatcherResult
import com.mapbox.navigation.core.trip.session.NavigationSessionState
import eu.ztsh.garmin.data.DataCache
import eu.ztsh.garmin.data.GarminMapper
import eu.ztsh.garmin.data.MapboxMapper
import java.io.IOException
import java.util.*
import java.util.concurrent.SynchronousQueue

@SuppressLint("MissingPermission")
class Garmin(
    val context: MainActivity,
    val device: BluetoothDevice,
    val adapter: BluetoothAdapter
) {

    private lateinit var connection: ConnectThread
    private lateinit var processing: ProcessingThread
    private val cache = DataCache()

    fun start() {
        connection = ConnectThread()
        connection.start()
    }

    fun close() {
        connection.close()
    }

    fun process(maneuver: Maneuver) {
        processing = ManeuverProcessingThread(maneuver)
        processing.start()
        processing.join()
    }

    fun process(location: LocationMatcherResult) {
        processing = LocationProcessingThread(location)
        processing.start()
        processing.join()
    }

    fun process(navigationSessionState: NavigationSessionState) {
        cache.update(navigationSessionState)
    }

    private inner class ManeuverProcessingThread(val maneuver: Maneuver) : ProcessingThread() {

        override fun run() {
            if (cache.hasChanged(maneuver)) {
                cache.update(maneuver)
                send(MapboxMapper.apply(maneuver))
            }
        }

    }

    private inner class LocationProcessingThread(val location: LocationMatcherResult) : ProcessingThread() {

        override fun run() {
            if (cache.hasChanged(location)) {
                cache.update(location)
                send(MapboxMapper.apply(location))
            }
        }

    }
        private fun setDistance(state: eu.ztsh.garmin.State) {
            connection.enqueue(intArrayOf(
                0x03,
                asDigit(state.distance / 1000), // position 1
                asDigit(state.distance / 100), // position 2
                asDigit(state.distance / 10), // position 3
                if ((state.distance * 10).toInt() == (state.distance.toInt() * 10)) 0x00 else 0xff, // comma
                asDigit(state.distance), // position 4
                state.unit.data // unit
            ))
        }

    private open inner class ProcessingThread : Thread() {

        fun send(incoming: eu.ztsh.garmin.data.State) {
            if (cache.hasChanged(incoming.distance)) {
                connection.enqueue(GarminMapper.setDistance(incoming))
            }
            if (cache.hasChanged(incoming.direction)) {
                connection.enqueue(GarminMapper.setDirection(incoming))
            }
            cache.update(incoming)
        }

        private fun asDigit(input: Double): Int {
            val number = input.toInt()
            if (number == 0) {
                return 0
            }
            val m = number % 10
            return if (m == 0) 10 else m
        }
    }

    private inner class ConnectThread : Thread() {

        private val queue: SynchronousQueue<IntArray> = SynchronousQueue()
        private var current: IntArray = intArrayOf()

        private val socket: BluetoothSocket? by lazy(LazyThreadSafetyMode.NONE) {
            context.checkBt()
            device.createRfcommSocketToServiceRecord(UUID.fromString("00001101-0000-1000-8000-00805F9B34FB"))
        }

        override fun run() {
            // Cancel discovery because it otherwise slows down the connection.
            context.checkBt()
            adapter.cancelDiscovery()
            try {
                socket?.connect()
                context.setConnectionStatus(true)
                sleep(3000)
                readAll()
                while (true) {
                    val newCurrent = queue.poll()
                    if (newCurrent == null) {
                        sleep(500)
                    } else {
                        current = newCurrent
                    }
                    send(current)
                }
            } catch (e: IOException) {
                Log.d(TAG, "Not connected", e)
                context.setConnectionStatus(false)
                while (true) {
                    // Just dequeue
                    // TODO: Add option to reconnect
                    queue.poll()
                    sleep(900)
                }
            }
        }

        // Closes the client socket and causes the thread to finish.
        fun close() {
            try {
                socket?.close()
            } catch (e: IOException) {
                Log.e(TAG, "Could not close the client socket", e)
            }
        }

        fun enqueue(data: IntArray) {
            queue.put(data)
        }

        private fun send(data: IntArray) {
            sendRaw(prepareData(data))
        }

        private fun readAll() {
            val buffer = ByteArray(64)
            val istr = socket!!.inputStream
            istr!!.let {
                if (it.available() > 0) {
                    it.read(buffer)
                }
            }
        }

        private fun sendRaw(buff: IntArray) {
            buff.forEach { socket!!.outputStream.write(it) }
            socket!!.outputStream.flush()
            sleep(2000)
            readAll()
        }

    }

    companion object {
        fun prepareData(input: IntArray): IntArray {
            val n = input.size
            var crc = (0xeb + n + n).toUInt()
            val chars = ArrayList<Int>()
            chars.add(0x10)
            chars.add(0x7b)
            chars.add((n + 0x06))
            if (n == 0xa)
                chars.add(0x10)
            chars.add(n)
            chars.add(0x00)
            chars.add(0x00)
            chars.add(0x00)
            chars.add(0x55)
            chars.add(0x15)
            for (char in input) {
                crc = (crc + char.toUInt())
                chars.add(char)
                if (char == 0x10)
                    chars.add(0x10)
            }
            chars.add((-(crc.toInt()) and 0xff))
            chars.add(0x10)
            chars.add(0x03)
            return chars.toIntArray()
        }

        private const val TAG = "GARMIN"
    }
}
